# coding: utf-8
# Taken from <https://github.com/rtfd/readthedocs.org/tree/2.8.3
#   readthedocs/search/parse_json.py>
#
# Copyright (c) 2010-2017 Read the Docs, Inc & contributors
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
from pyquery import PyQuery
import logging
import codecs
import fnmatch
import json
import os


log = logging.getLogger(__name__)


def process_all_json_files(build_dir):
    """Return a list of pages to index"""
    html_files = []
    for root, _, files in os.walk(build_dir):
        for filename in fnmatch.filter(files, '*.fjson'):
            if filename in ['search.fjson', 'genindex.fjson',
                            'py-modindex.fjson']:
                continue
            html_files.append(os.path.join(root, filename))
    page_list = []
    for filename in html_files:
        try:
            result = process_file(filename)
            if result:
                page_list.append(result)
        # we're unsure which exceptions can be raised
        except:  # noqa
            pass
    return page_list


def process_file(filename):
    """Read a file from disk and parse it into a structured dict."""
    try:
        with codecs.open(filename, encoding='utf-8', mode='r') as f:
            file_contents = f.read()
    except IOError:
        log.info('Unable to read file: %s', filename)
        return None
    data = json.loads(file_contents)
    title = ''
    body_content = ''
    if 'current_page_name' in data:
        path = data['current_page_name']
    else:
        log.info('Unable to index file due to no name %s', filename)
        return None
    if 'body' in data and data['body']:
        body = PyQuery(data['body'])
        body_content = body.text().replace(u'¶', '')
    else:
        log.info('Unable to index content for: %s', filename)
    if 'title' in data:
        title = data['title']
        if title.startswith('<'):
            title = PyQuery(data['title']).text()
    else:
        log.info('Unable to index title for: %s', filename)

    return {'headers': process_headers(data, filename),
            'content': body_content, 'path': path, 'title': title}


def process_headers(data, filename):
    """Read headers from toc data."""
    headers = []
    if 'toc' in data:
        for element in PyQuery(data['toc'])('a'):
            headers.append(recurse_while_none(element))
        if None in headers:
            log.info('Unable to index file headers for: %s', filename)
    return headers


def recurse_while_none(element):
    if element.text is None:
        return recurse_while_none(element.getchildren()[0])
    return element.text



# Inspired by <https://github.com/rtfd/readthedocs.org/tree/2.8.3/
#   readthedocs/search/indexes.py>
import click
import elasticsearch
import elasticsearch.helpers
import hashlib
import json
import os.path
import sphinx_elasticsearch.parse_json
import sys


@click.group()
def cli():
    pass


@cli.command()
@click.option('--es-url', default='http://localhost:9200')
@click.option('--index', default='docs')
@click.option('--type', default='page')
@click.option('--project-name', required=True)
@click.option('--commit')
@click.argument('build-dir')
def index(es_url, index, type, project_name, commit, build_dir):
    if not os.path.exists(build_dir):
        raise IOError('Directory does not exist: %s' % build_dir)

    es = elasticsearch.Elasticsearch([es_url])

    pages = sphinx_elasticsearch.parse_json.process_all_json_files(build_dir)
    to_index = []
    for page in pages:
        page_id = hashlib.md5((
            project_name + '-' + page['path']).encode('utf-8')).hexdigest()
        doc = {
            'id': page_id,
            'project': project_name,
            'path': page['path'],
            'title': page['title'],
            'headers': page['headers'],
            'content': page['content'],
            'commit': commit,
        }
        to_index.append({
            '_index': index,
            '_type': type,
            '_id': page_id,
            '_source': doc,
        })
    elasticsearch.helpers.bulk(es, to_index)

    if commit:
        es.delete_by_query(body={'query': {'bool': {
            'must': [{'term': {'project': project_name}}],
            'must_not': [{'term': {'commit': commit}}],
        }}}, index=index, doc_type=type)


@cli.command()
@click.option('--es-url', default='http://localhost:9200')
@click.option('--project-name', default=None)
@click.option('--index', default='docs')
@click.argument('query')
def search(es_url, index, query, project_name):
    # This is an example for reference, for a complete implementation see
    # <https://github.com/ZeitOnline/sphinx_zon_theme>
    es = elasticsearch.Elasticsearch([es_url])
    body = {'query': {'bool': {'should': [
        {'match_phrase': {'title': {'query': query, 'boost': 10, 'slop': 2}}},
        {'match_phrase': {'headers': {'query': query, 'boost': 5, 'slop': 3}}},
        {'match_phrase': {'content': {'query': query, 'slop': 5}}},
    ]}},
        'highlight': {'fields': {
            'title': {},
            'headers': {},
            'content': {},
        }},
        '_source': ['title', 'project', 'path'],
        'size': 25,
    }

    if project_name:
        body['query']['bool']['filter'] = [
            {'term': {'project': project_name}}]

    json.dump(es.search(index=index, body=body), sys.stdout, indent=2)


@cli.command()
@click.option('--es-url', default='http://localhost:9200')
@click.option('--index', default='docs')
@click.option('--replicas', required=True)
@click.option('--shards', required=True)
def create_index(es_url, index, replicas, shards):
    # This is mostly for convenience so you can get started out-of-the-box;
    # if you have specific elasticsearch requirements, you'll be better off
    # creating the index yourself externally.
    es = elasticsearch.Elasticsearch([es_url])
    es.indices.create(index=index, body={'settings': {
        'number_of_replicas': replicas,
        'number_of_shards': shards,
    }})


@cli.command()
@click.option('--es-url', default='http://localhost:9200')
@click.option('--index', default='docs')
@click.option('--type', default='page')
def put_mapping(es_url, index, type):
    # This is the simplest possible mapping for the fields we need; you'll
    # probably want to at least add stemmers for your documentation language.
    # And like for `create_index`, you'll probably be better off
    # managing the mapping yourself externally.
    es = elasticsearch.Elasticsearch([es_url])
    es.indices.put_mapping(type, {type: {
        # Disable _all field to reduce index size.
        '_all': {'enabled': False},
        'properties': {
            'id': {'type': 'keyword'},
            'project': {'type': 'keyword'},
            'path': {'type': 'keyword'},
            'commit': {'type': 'keyword'},

            'title': {'type': 'text'},
            'headers': {'type': 'text'},
            'content': {'type': 'text'},
        }
    }}, index)



from setuptools import setup, find_packages


setup(
    name='sphinx_elasticsearch',
    version='1.0.0',
    author='Zeit Online',
    author_email='zon-backend@zeit.de',
    url='https://github.com/zeitonline/sphinx_elasticsearch',
    description="Indexes documentation build via Sphinx into elasticsearch",
    long_description='\n\n'.join(
        open(x).read() for x in ['README.rst', 'CHANGES.txt']),
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    zip_safe=False,
    license='BSD',
    classifiers=[
        'Programming Language :: Python :: 3',
    ],
    install_requires=[
        'click',
        'elasticsearch',
        'pyquery',
        'setuptools',
    ],
    extras_require={'test': [
        'pytest',
    ]},
    entry_points={
        'console_scripts': [
            'sphinx-elasticsearch = sphinx_elasticsearch.index:cli',
        ]
    }
)
